/* pal
 *
 * Copyright (C) 2003, Scott Kuhl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <readline/readline.h>

#include "main.h"
#include "output.h"
#include "event.h"
#include <stdlib.h>    /* for exit() */

char* default_text;

char* no_match()
{ return NULL; }

/* prompt for required input */
/* caller is responsible for freeing the prompt and the read line */
gchar* get_line(char* prompt)
{
    char *line = NULL;
    do
    {
	if(line != NULL)
	    g_free(line);
	line = readline(prompt);

	if(line == NULL) /* line is null with ^D */
	    exit(0);

	g_strstrip(line);

    }
    while(*line == '\0');

    return line;
}

gboolean get_line_y_n(char* prompt)
{
    gchar *s = NULL;

    do
    {
	g_free(s);
	s = get_line(prompt);
    }
    while(*s != 'y' && *s != 'Y' && *s != 'n' && *s != 'N');

    if(*s == 'y' || *s == 'Y')
	return TRUE;
    return FALSE;

}

void print_default_text()
{
    rl_insert_text(default_text);
    rl_redisplay();
}

/* returned string should be freed */
void pal_add_suffix(gint number, gchar* suffix)
{
    number = number % 10;
    switch(number)
    {
	case 1:  g_sprintf(suffix, "%s", "st"); return;
	case 2:  g_sprintf(suffix, "%s", "nd"); return;
	case 3:  g_sprintf(suffix, "%s", "rd"); return;
	default: g_sprintf(suffix, "%s", "th"); return;
    }

}

gchar* pal_add_get_range()
{
    g_print("\n");
    pal_output_fg(BRIGHT, GREEN, "> ");
    g_print(_("Does this event have starting and ending dates? "));

    if(get_line_y_n(_("[y/n]: ")))
    {
	GDate* d1 = NULL;
	GDate* d2 = NULL;
	gchar buf[1024];
	gchar* s;
	gboolean bad_range = FALSE;

	do
	{
	    bad_range = FALSE;

	    while(d1 == NULL)
	    {
		s = get_line("Start date: ");
		d1 = get_query_date(s, FALSE);
		g_free(s);
	    }

	    while(d2 == NULL)
	    {
		s = get_line("End date: ");
		d2 = get_query_date(s, FALSE);
		g_free(s);
	    }

	    if(g_date_days_between(d1,d2) < 1)
	    {
		pal_output_error("ERROR: The end date is the same as or before the start date.\n");
		d1 = NULL;
		d2 = NULL;
		bad_range = TRUE;
		continue;
	    }


	    g_date_strftime(buf, 1024, "Start date: %a %e %b %Y\n", d1);
	    g_print(buf);
	    g_date_strftime(buf, 1024, "End date: %a %e %b %Y\n", d2);
	    g_print(buf);

	    sprintf(buf, "%s ", _("Accept? [y/n]:"));

	} while(bad_range || !get_line_y_n(buf));

	s = g_strconcat(":", get_key(d1), ":", get_key(d2), NULL);
	g_date_free(d1);
	g_date_free(d2);
	return s;

    }
    else return g_strdup("");

}


/* reuturned string should be freed */
gchar* pal_add_get_recur(GDate* date)
{
    gchar* key;
    gchar* s;

    g_print("\n");
    pal_output_fg(BRIGHT, GREEN, "> ");
    g_print(_("Is the event recurring? "));

    key = get_key(date);

    if(get_line_y_n(_("[y/n]: ")))
    {
	gchar* selection;
	gchar buf[128];
	gchar suffix[3];
	gboolean last_weekday = FALSE;

	pal_output_fg(BRIGHT, GREEN, "> ");
	g_print(_("Select how often this event occurs\n"));

	g_date_strftime(buf, 128, "%B", date);
	pal_add_suffix(g_date_get_day(date), suffix);


	pal_output_attr(BRIGHT, " 1 ");
	g_print(_("- Daily\n"));

	pal_output_attr(BRIGHT, " 2 ");
	g_date_strftime(buf, 128, "%A", date);
	g_print(_("- Weekly (every %s)\n"), buf);

	pal_output_attr(BRIGHT, " 3 ");
	pal_add_suffix(g_date_get_day(date), suffix);
	g_print(_("- Monthly (on the %d%s of every month)\n"),
		g_date_get_day(date), suffix);

	pal_output_attr(BRIGHT, " 4 ");
	g_print(_("- Annually (every %s %d%s)\n"), buf,
		g_date_get_day(date), suffix);

	pal_output_attr(BRIGHT, " 5 ");
	pal_add_suffix(get_nth_day(date), suffix);
	g_date_strftime(buf, 128, "%A of every %B", date);
	g_print(_("- The %d%s %s.\n"), get_nth_day(date),
		suffix, buf);


	if(last_weekday_of_month(date))
	{
	    last_weekday = TRUE;
	    pal_output_attr(BRIGHT, " 6 ");
	    g_date_strftime(buf, 128, "%A of every %B", date);
	    g_print(_("- The last %s.\n"), buf);
	}

	g_print("\n");

	do
	{
	    if(last_weekday)
		selection = get_line(_("Select type [1--6]: "));
	    else
		selection = get_line(_("Select type [1--5]: "));


	    if(g_ascii_strcasecmp(selection, "1") == 0)
	    {
		sprintf(key, "DAILY");
		break;
	    }
	    else if(g_ascii_strcasecmp(selection, "2") == 0)
	    {
		switch(g_date_get_weekday(date))
		{
		    case 1: sprintf(key, "MON"); break;
		    case 2: sprintf(key, "TUE"); break;
		    case 3: sprintf(key, "WED"); break;
		    case 4: sprintf(key, "THU"); break;
		    case 5: sprintf(key, "FRI"); break;
		    case 6: sprintf(key, "SAT"); break;
		    case 7: sprintf(key, "SUN"); break;
		    default: key = NULL; /* impossible...i think */
		}
		break;
	    }
	    else if(g_ascii_strcasecmp(selection, "3") == 0)
	    {
		key[0] = '0';
		key[1] = '0';
		key[2] = '0';
		key[3] = '0';
		key[4] = '0';
		key[5] = '0';
		break;
	    }
	    else if(g_ascii_strcasecmp(selection, "4") == 0)
	    {
		key[0] = '0';
		key[1] = '0';
		key[2] = '0';
		key[3] = '0';
		break;
	    }
	    else if(g_ascii_strcasecmp(selection, "5") == 0)
	    {
		int weekday = g_date_get_weekday(date);
		int month = g_date_get_month(date);
		int n = get_nth_day(date);

		/* convert weekday to friendly weekday
		   from: 1(mon) -> 7(sun)
		   to:   1(sun) -> 7(sat) */
		if(weekday == 7)
		    weekday = 1;
		else
		    weekday++;

		g_sprintf(key, "*%02d%01d%01d", month, n, weekday);
		break;
	    }
	    else if(g_ascii_strcasecmp(selection, "6") == 0 && last_weekday)
	    {
		int weekday = g_date_get_weekday(date);
		int month = g_date_get_month(date);

		/* convert weekday to friendly weekday
		   from: 1(mon) -> 7(sun)
		   to:   1(sun) -> 7(sat) */
		if(weekday == 7)
		    weekday = 1;
		else
		    weekday++;

		g_sprintf(key, "*%02dL%01d", month, weekday);
		break;
	    }

	}while(1);


	s = g_strconcat(key, pal_add_get_range(), NULL);
	return s;


    }
    else
	return get_key(date);
}


/* prompts for a date */
gchar* pal_add_get_date()
{
    gchar* s = NULL;
    GDate* d = NULL;

    g_print("\n");

    pal_output_fg(BRIGHT, GREEN, "> ");
    g_print(_("When is the event? (formats: yyyymmdd, Jan 1 2000, 1 Jan 2000, 4 days away)\n"));
    g_print(_("Use \"TODO\" to make an event that always occurs on the current date.\n"));
    g_print(_("If the event is recurring, select one of the days the event occurs.\n"));

    do
    {
	s = get_line("Date for event: ");
	d = get_query_date(s, FALSE);

	if(d != NULL)
	{
	    gchar buf[1024];

	    g_print("\n");

	    pal_output_fg(BRIGHT, GREEN, "> ");
	    g_print(_("Other events on the date you selected:\n"));

	    g_print("\n");
	    pal_output_date(d, TRUE);
	    g_print("\n");

	    pal_output_fg(BRIGHT, GREEN, "> ");
	    g_print(_("Is this the date you want to add an event to?\n"));

	    g_date_strftime(buf, 1024, _("%a %e %b %Y - Accept? [y/n]: "), d);

	    if(get_line_y_n(buf))
	    {
		s = pal_add_get_recur(d);
		g_date_free(d);
		return s;
	    }
	}

	else
	{
	    if(g_ascii_strcasecmp(s, "todo") == 0)
		return g_strdup("TODO");
	}

	if(d != NULL) g_date_free(d);
	if(s != NULL) g_free(s);

    } while(1);



    /* impossible */
    return NULL;
}


gchar* pal_add_get_desc()
{
    char* desc = NULL;

    g_print("\n");
    pal_output_fg(BRIGHT, GREEN, "> ");
    g_print(_("What is the description of this event?\n"));

    do{
	if(desc != NULL)
	{
	    default_text = desc;
	    rl_pre_input_hook = (rl_hook_func_t*) print_default_text;
	}

	desc = get_line(_("Description: "));
	rl_pre_input_hook = NULL;
    }
    while(!get_line_y_n(_("Is this description correct? [y/n]: ")));

    return desc;
}


/* prompts for a file name */
gchar* pal_add_get_file()
{
    char* filename;
    gboolean prompt_again = FALSE;

    pal_output_fg(BRIGHT, GREEN, "> ");
    g_print(_("Calendar file (usually ending with \".pal\") to add event to:\n"));

    /* get the filename */
    do
    {
	prompt_again = FALSE;
	default_text = g_strconcat(g_get_home_dir(), "/.pal/", NULL);
	rl_pre_input_hook = (rl_hook_func_t*) print_default_text;
	filename = get_line("Filename: ");
	rl_pre_input_hook = NULL;

	/* replace ~/ or ~ with home dir  */
	if(*filename == '~')
	{
	    char* tmp;
	    char* tmp_orig;

	    tmp = g_strconcat(filename, NULL);
	    tmp_orig = tmp;

	    g_free(filename);
	    tmp++;
	    if(*tmp == '/')
		tmp++;
	    filename = g_strconcat(g_get_home_dir(), "/", tmp, NULL);
	    g_free(tmp_orig);
	}

	/* check if file exists */
	if(g_file_test(filename, G_FILE_TEST_EXISTS))
	{
	    /* make sure it aint a directory */
	    if(g_file_test(filename, G_FILE_TEST_IS_DIR))
	    {
		pal_output_error(_("ERROR: %s is a directory.\n"), filename);
		prompt_again = TRUE;
	    }
	}
	else
	{
	    /* ask to create the file if it doesn't exist */
	    pal_output_error(_("WARNING: %s does not exist.\n"), filename);

	    if(!get_line_y_n(_("Create? [y/n]: ")))
	       prompt_again = TRUE;
	    else
	    {
		/* create the file */
		FILE*  file = fopen(filename, "w");
		if(file == NULL)
		{
		    pal_output_error(_("ERROR: Can't create %s.\n"), filename);
		    prompt_again = TRUE;
		}
		else
		{
		    gchar *markers;
		    gchar *event_type;
		    gchar *top_line;

		    g_print("\n");
		    pal_output_fg(BRIGHT, GREEN, "> ");
		    g_print(_("Information for %s:\n"), filename);
		    do
			markers = get_line(_("2 character marker for calendar: "));
		    while(*markers == '\0' || *(markers+1) == '\0');

		    event_type = get_line(_("Calendar title: "));

		    top_line = g_strconcat(markers, " ", event_type, "\n", NULL);
		    fputs(top_line, file);

		    g_free(top_line);
		    g_free(event_type);
		    g_free(markers);

		    fclose(file);
		}
	    }

	}

	if(prompt_again)
	    g_free(filename);

    }while (prompt_again);

    return filename;
}

void pal_add_write_file(gchar* filename, gchar* key, gchar* desc)
{
    FILE *file = NULL;
    gchar* write_line = NULL;
    gboolean no_newline = FALSE;

    /* check for newline at end of file */
    do
    {
	file = fopen(filename, "r");
	if(file == NULL)
	{
	    pal_output_error(_("ERROR: Can't read to file %s.\n"), filename);
	    if(!get_line_y_n(_("Try again? [y/n]: ")))
		return;
	}
    } while(file == NULL);

    fseek(file, -1, SEEK_END);
    if(fgetc(file)!= '\n')
	no_newline = TRUE;
    fclose(file);

    /* write the new event out to that file */
    do
    {
	file = fopen(filename, "a");
	if(file == NULL)
	{
	    pal_output_error(_("ERROR: Can't write to file %s.\n"), filename);
	    if(!get_line_y_n(_("Try again? [y/n]: ")))
		return;
	}
    } while(file == NULL);

    /* put a newline at end of file if one isn't there already */
    if(no_newline)
	fputc('\n', file);

    {
	write_line = g_strconcat(key, " ", desc, "\n", NULL);
	fputs(write_line, file);
	g_print("\n");
	pal_output_fg(BRIGHT, GREEN, ">>> ");
	g_print(_("Wrote \"%s %s\" to %s.\n"), key, desc, filename);
	g_free(key);
	g_free(write_line);
	fclose(file);
    }
}


void pal_add_event()
{
    gchar* filename;
    gchar* description;
    gchar* key;

    gboolean first_time_flag = TRUE;

    pal_output_fg(BRIGHT, GREEN, "* * * ");
    pal_output_attr(BRIGHT, _("Add an event"));
    pal_output_fg(BRIGHT, GREEN, " * * *\n");

    g_print(_("Press Control+c at any time to cancel.\n\n"));


    do
    {
	if(!first_time_flag)
	    g_print("\n");

	first_time_flag = FALSE;

	rl_completion_entry_function = rl_filename_completion_function;

	filename = pal_add_get_file();

	/* no more completion necessary */
	rl_completion_entry_function = (rl_compentry_func_t*) no_match;

	key = pal_add_get_date();
	description = pal_add_get_desc();

	if(!is_valid_key(key))
	{
	    pal_output_error("INTERNAL ERROR: Please report this error message along with\n");
	    pal_output_error("                the input that generated it.\n");
	    pal_output_error("INVALID KEY: %s\n", key);

	}
	else
	    pal_add_write_file(filename, key, description);

	g_free(filename);
	g_free(description);
	g_free(key);

    } while(get_line_y_n(_("Add another event? [y/n]: ")));
}



