/* pal
 *
 * Copyright (C) 2003, Scott Kuhl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <time.h>
#include <string.h>

/* mkdir */
#include <sys/stat.h>
#include <sys/types.h>


#include "main.h"
#include "output.h"
#include "event.h"
#include "colorize.h"

/* checks if events in the format yyyymmdd can be expunged */
gboolean should_be_expunged(const gchar* key, const PalEvent* pal_event)
{
    GDate* today = g_date_new();
    GDate* event_day = get_date(key);
    g_date_set_time(today, time(NULL));

    /* if not a yyyymmdd (ie, not recurring) */
    if(event_day == NULL)
    {
	/* recurring event with end_date */
	if(pal_event->end_date != NULL &&
	   g_date_days_between(today, pal_event->end_date) <= -1*settings->expunge)
	{
	    g_date_free(today);
	    return TRUE;
	}

	g_date_free(today);
	return FALSE;
    }

    /* if it is a yyyymmdd event (ie one-time event) */
    if(g_date_days_between(today, event_day) <= -1*settings->expunge)
    {
	g_date_free(today);
	g_date_free(event_day);
	return TRUE;
    }

    g_date_free(today);
    g_date_free(event_day);
    return FALSE;


}




/* loads a pal calendar file, returns the number of events loaded */
gint load_file(gchar* filename, FILE* file, gint filecount, gboolean hide, int color)
{
    gchar s[2048];
    gint eventcount = 0;

    GList* days_events = NULL;
    PalEvent* pal_event = NULL;

    FILE *out_file = NULL;
    gchar* out_filename = NULL;

    gchar *key = NULL;

    gchar start, end, c;
    gchar type[1024];
    gboolean do_expunge = FALSE;

    g_strstrip(filename);
    out_filename = g_strconcat(filename, ".paltmp", NULL);

    /* if -x is used and the file isn't a global calendar, expunge */
    if(settings->expunge > 0 && !g_str_has_prefix(out_filename, PREFIX "/share/pal"))
    {
	out_file = fopen(out_filename, "w");
	if(out_file == NULL)
	{
	    pal_output_error(_("ERROR: Can't write file: %s\n"), out_filename);
	    pal_output_error(_("       File will not be expunged: %s\n"), filename);
	}
	else
	    do_expunge = TRUE;
    }

    /* read in comments at top of file */
    do
    {
	gchar* orig_string;
	fgets(s, 2048, file);
	orig_string = g_strdup(s);
	g_strstrip(s);

	if(do_expunge)
	    fputs(orig_string, out_file);

	g_free(orig_string);

    } while(*s == '#' || *s == '\0');

    /* we just read the "first" line */

    sscanf(s, "%c%c%c%s", &start, &end, &c, type);

    if(c != ' ' && c != '\t') /* there should be white space here */
    {
	gchar* file = g_path_get_basename(filename);
	pal_output_error(_("ERROR: First line is improperly formatted.\n"));
	pal_output_error(  "       %s: %s\n", _("FILE"), file);
	g_free(file);
	pal_output_error(  "       %s: %s\n", _("LINE"), s);
    }

    g_strstrip(type);

    /* read in the events in the file */
    while(fgets(s, 2048, file) != NULL)
    {
	gchar date_string[128];
	gchar* text_string;
	gchar* tmp;


	tmp = strdup(s);
	g_strstrip(tmp);

	/* don't need to do anything if we just read a blank line or comment */
	if(*tmp == '\0' || *tmp == '#')
	{
	    /* copy blank line */
	    if(do_expunge)
		fputs(s, out_file);

	    g_free(tmp);
	    continue;
	}
	g_free(tmp); tmp = NULL;

	/* first word is the date string */
	sscanf(s, "%s", date_string);

	/* the rest is the descriptive text */
	text_string = g_strdup(s+strlen(date_string));


	/* around about way of making date_string uppercase.  The keys
	   in hashtable must be uppercase, but pal's .pal files are
	   case insensitive. */
	g_strstrip(date_string);
	tmp = g_ascii_strup(date_string, -1);
	sscanf(tmp, "%s", date_string);
	g_free(tmp);

	g_strstrip(text_string);
	eventcount++;

	/* check for a valid date_string */
	if(!is_valid_key(date_string))
	{
	    gchar* file = g_path_get_basename(filename);
	    pal_output_error(_("ERROR: Invalid date string.\n"));
	    pal_output_error(  "       %s: %s\n", _("FILE"), file);
	    pal_output_error(  "       %s: %s\n", _("LINE"), s);
	    g_free(file);

	    /* copy bad line */
	    if(do_expunge)
		fputs(s, out_file);

	    g_free(text_string);
	    continue;
	}

	/* require a description for a event */
	if(strlen(text_string) == 0)
	{
	    gchar* file = g_path_get_basename(filename);
	    pal_output_error(_("ERROR: Event description missing.\n"));
	    pal_output_error(  "       %s: %s\n", _("FILE"), file);
	    pal_output_error(  "       %s: %s\n", _("LINE"), s);
	    g_free(file);

	    /* copy bad line */
	    if(do_expunge)
		fputs(s, out_file);

	    g_free(text_string);
	    continue;
	}

	pal_event = g_new0(PalEvent,1);
	pal_event->text = g_strdup(text_string);
	pal_event->hide = hide;
	pal_event->start = start;
	pal_event->end = end;
	pal_event->type = g_strdup(type);
	pal_event->file_num = filecount;
	pal_event->color = color;

	key = pal_event_fill_dates(pal_event, date_string);

	if(do_expunge)
	{
	    if(should_be_expunged(key, pal_event))
	    {
		if(settings->verbose)
		    g_print("%s: %s", _("Expunged"), s);
		eventcount--;

		g_free(key);
		g_free(text_string);
		pal_event_free(pal_event);
		continue;
	    }
	    else
		fputs(s, out_file);
	}


	/* if no list exists for that key, make new list */
	if(g_hash_table_lookup(ht, key) == NULL)
	{
	    days_events = NULL;
	    days_events = g_list_append(days_events, pal_event);
	    g_hash_table_insert(ht,g_strdup(key),days_events);
	}
	else /* else, append to current list */
	{
	    days_events = g_hash_table_lookup(ht,key);
	    days_events = g_list_append(days_events,pal_event);
	}

	g_free(key);
	g_free(text_string);
    }


    if(do_expunge)
    {
	fclose(out_file);
	if(rename(out_filename, filename) != 0)
	    pal_output_error(_("ERROR: Can't rename %s to $s\n"), out_filename, filename);
    }

    g_free(out_filename);

    return eventcount;
}


/* gets the pal file to load.  From "file", it looks for a pal
 * calendar file locally or globally to load and puts the path of that
 * file in pal_file.  It returns true if successful. */
gboolean get_file_to_load(gchar* file, gchar* pal_file, gboolean show_error)
{
    if(g_path_is_absolute(file))
    {
	if(! g_file_test(file, G_FILE_TEST_EXISTS) || g_file_test(file, G_FILE_TEST_IS_DIR))
	{
	    if(show_error)
		pal_output_error(_("ERROR: File doesn't exist: %s\n"), file);
	    return FALSE;
	}
	else
	    g_sprintf(pal_file, file);
    }

    else
    {
	/* if a relative path, try looking in the path found in settings->conf_file */
	gchar* dirname = g_path_get_dirname(settings->conf_file);
	g_sprintf(pal_file, "%s/%s", dirname, file);

	/* if that doesn't work, try looking in PREFIX/share/pal */
	if(! g_file_test(pal_file, G_FILE_TEST_EXISTS) || g_file_test(pal_file, G_FILE_TEST_IS_DIR))
	    g_sprintf(pal_file, PREFIX "/share/pal/%s", file);

	/* if that doesn't work, print message */
	if(! g_file_test(pal_file, G_FILE_TEST_EXISTS) || g_file_test(pal_file, G_FILE_TEST_IS_DIR))
	{
	    gchar other_file[2048];
	    g_sprintf(other_file, "%s/%s", dirname, file);
	    if(show_error)
		pal_output_error(_("ERROR: Can't find file.  I tried %s and %s.\n"),
				 other_file, pal_file);

	    g_free(dirname);
	    return FALSE;
	}

	g_free(dirname);
    }

    return TRUE;
}


/* opens a file for reading.  Prints error if can't read file.
 * Returns NULL if can't read, returns a FILE pointer if reading file.
 * fclose() should be called on the returned pointer when done reading
 * the file. */
FILE* get_file_handle(gchar* filename, gboolean show_error)
{
    FILE* file = fopen(filename, "r");

    if(settings->verbose)
	g_print(_("Reading: %s\n"), filename);

    if(file == NULL && show_error)
	pal_output_error(_("ERROR: Can't read file: %s\n"), filename);

    return file;
}


/* loads calendar files and settings from a pal.conf file */
GHashTable* load_files()
{
    gchar s[2048];
    gchar text[2048];
    FILE* file = NULL;
    guint filecount=0, eventcount=0;

    ht = g_hash_table_new(g_str_hash,g_str_equal);

    if(settings->verbose)
    {
	if(settings->expunge >= 0)
	    g_print(_("Looking for data to expunge.\n"));
    }

    file = get_file_handle(settings->conf_file, FALSE);

    if(file == NULL)
    {
	if(settings->specified_conf_file)
	{
	    pal_output_error(_("ERROR: Can't open file: %s\n"), settings->conf_file);
	    return ht;
	}
	else /* didn't specify conf file, and couldn't find conf file: create a default one */
	{
	    FILE* out_file;
	    gchar* out_dirname;
	    gchar* out_path;
	    gint c;

	    out_dirname = g_strconcat(g_get_home_dir(), "/.pal", NULL);
	    out_path    = g_strconcat(out_dirname, "/pal.conf", NULL);

	    pal_output_error(_("NOTE: Creating %s\n"), out_path);
	    pal_output_error(_("NOTE: Edit ~/.pal/pal.conf to change how and if certain events are displayed.\n"),
			     out_path);

	    /* create directory if it doesn't exist */
	    if(!g_file_test(out_dirname, G_FILE_TEST_IS_DIR))
	    {
		if(mkdir(out_dirname, 0755) != 0)
		{
		    pal_output_error(_("ERROR: Can't create directory: %s\n"), out_dirname);
		    return ht;
		}
	    }

	    /* attempt to copy PREFIX/share/pal/pal.conf to ~/.pal/pal.conf */
	    file = fopen(PREFIX "/share/pal/pal.conf", "r");

	    if(file == NULL)
	    {
		pal_output_error(_("ERROR: Can't open file: " PREFIX "/share/pal/pal.conf\n"));
		pal_output_error(_("ERROR: This indicates an improper installation.\n"));
		return ht;
	    }

	    out_file = fopen(out_path, "w");
	    if(out_file == NULL)
	    {
		pal_output_error(_("ERROR: Can't create/write file: %s\n"), out_file);
		return ht;
	    }

	    /* do copy */
	    c = fgetc(file);
	    while(c != EOF)
	    {
		fputc(c,out_file);
		c=fgetc(file);
	    }

	    fclose(out_file);
	    fclose(file);

	    g_free(out_dirname);
	    g_free(out_path);

	    /* open file to read it as usual */
	    file = fopen(settings->conf_file, "r");
	}

    } /* done opening/creating file */

    /* if using -p, load that .pal file now. */
    if(settings->pal_file != NULL)
    {
	gchar pal_file[16384];
	FILE* pal_file_handle = NULL;

	if(!get_file_to_load(settings->pal_file, pal_file, FALSE))
	    g_sprintf(pal_file, settings->pal_file);

	pal_file_handle = get_file_handle(pal_file, TRUE);
	if(pal_file_handle != NULL)
	{
	    eventcount += load_file(pal_file, pal_file_handle, filecount, FALSE, -1);
	    fclose(pal_file_handle);
	    filecount++;
	}
    }


    while(fgets(s, 2048, file) != NULL)
    {
	gchar pal_file[16384];
	gchar color[16384];
	gint int_color = -1;
	gint i,j;
	color[0] = '\0';
	g_strstrip(s);

	if(sscanf(s, "file %s (%[a-z])\n", text, color) == 2 ||
	    sscanf(s, "file %s\n", text) == 1)
	{
	    FILE* pal_file_handle = NULL;
	    gboolean hide = FALSE;

	    /* skip this line if we're using -p */
	    if(settings->pal_file != NULL)
		continue;

	    if(sscanf(s, "file_hide %s (%[a-z])\n", text, color) == 1)
		hide = TRUE;
	    else if(sscanf(s, "file_hide %s\n", text) == 2)
		hide = TRUE;

	    if(color[0] != '\0')
	    {
		if(int_color_of(color) != -1)
		    int_color = int_color_of(color);

		if(int_color == -1)
		{
		    pal_output_error("ERROR: Not a valid color \"%s\" in %s\n",
				     color, settings->conf_file);
		    pal_output_error("       Valid colors: black, red, green, yellow, blue, magenta, cyan, white\n");
		}
	    }

	    if(get_file_to_load(text,pal_file, TRUE))
	    {
		pal_file_handle = get_file_handle(pal_file, TRUE);
		if(pal_file_handle != NULL)
		{
		    /* assign events that are the "default" color to
		     * have a color of -1 the output code will apply
		     * the default color to events (since we might not
		     * have read in what the default color is yet. */
		    eventcount += load_file(pal_file, pal_file_handle, filecount, hide, int_color);
		    fclose(pal_file_handle);
		    filecount++;
		}
	    }
	}

	else if(sscanf(s, "date_fmt %s",text) == 1)
	{
	    g_free(settings->date_fmt);
	    settings->date_fmt = g_strdup(g_strstrip(&s[9]));
	}
	else if(strcmp(s, "week_start_monday") == 0)
	    settings->week_start_monday = TRUE;
	else if(strcmp(s, "remind_reverse") == 0)
	{
	    settings->reverse_order = TRUE;
	    pal_output_error("WARNING: 'remind_reverse' in %s is deprecated.\n   Use 'reverse_order' instead.\n",
			     settings->conf_file);
	}
	else if(strcmp(s, "reverse_order") == 0)
	    settings->reverse_order = TRUE;
	else if(strcmp(s, "cal_on_bottom") == 0)
	    settings->cal_on_bottom = TRUE;
	else if(strcmp(s, "no_columns") == 0)
	    settings->no_columns = TRUE;
	else if(strcmp(s, "hide_remind_titles") == 0)
	{
	    settings->hide_event_type = TRUE;
	    pal_output_error("WARNING: 'hide_remind_titles' in %s is deprecated.\n   Use 'hide_event_type' instead.\n",
			     settings->conf_file);

	}
	else if(strcmp(s, "hide_event_type") == 0)
	    settings->hide_event_type = TRUE;
	else if(strcmp(s, "compact_list") == 0)
	    settings->compact_list = TRUE;
	else if(sscanf(s, "compact_date_fmt %s",text) == 1)
	{
	    g_free(settings->compact_date_fmt);
	    settings->compact_date_fmt = g_strdup(g_strstrip(&s[17]));
	}
	else if(sscanf(s, "event_color %s",text) == 1)
	{
	    int ec = int_color_of(text);
	    if(ec == -1)
		pal_output_error(_("ERROR: \"%s\" is not a valid color in %s\n"),
				 text, settings->conf_file);
	    else
		settings->event_color = ec;
	}
	else if(sscanf(s, "default_range %d-%d", &i, &j) == 2)
	{
	    /* check if -r was used */
	    if(!settings->range_arg)
	    {
		settings->range_neg_days = i;
		settings->range_days = j;
	    }
	}
	else if(sscanf(s, "default_range %d", &i) == 1)
	{
	    if(!settings->range_arg)
		settings->range_days = i;
	}

        /* ignore empty lines or comments */
	else if(*s != '#' && *s != '\0')
	{
	    pal_output_error("%s %s:\n       %s\n", _("ERROR: Invalid line in"),
			     settings->conf_file, s);
	}


    }
    fclose(file);
    if(settings->verbose)
	g_print("%s (%d %s, %d %s).\n\n",
		_("Done reading data"), eventcount, _("events"), filecount, _("files"));
    return ht;
}
