/* pal
 *
 * Copyright (C) 2003, Scott Kuhl
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdlib.h>    /* for exit() */
#include <sys/ioctl.h> /* get # columns for terminal */
#include <time.h>
#include <string.h>
#include <locale.h>
#include <sys/types.h> /* FreeBSD, regex.h needs this */
#include <regex.h>     /* regular expressions */

#include "colorize.h"
#include "output.h"
#include "main.h"
#include "input.h"
#include "event.h"
#include "add.h"
#include "html.h"

void view_search(GDate* date, gint window)
{
    regex_t preg;
    gchar start_date[128];
    gchar end_date[128];
    gchar* search = settings->search_string;
    gint i,j;
    gint search_hits = 0;

    g_date_strftime(start_date, 128, settings->date_fmt, date);
    g_date_add_days(date, window-1);
    g_date_strftime(end_date, 128, settings->date_fmt, date);
    g_date_subtract_days(date, window-1);

    pal_output_attr(BRIGHT, _("[ Begin search results: %s ]\n[ From %s to %s inclusive ]\n\n"),
	    search, start_date, end_date);

    if(settings->reverse_order)
	g_date_add_days(date, window-1);

    regcomp(&preg, search, REG_ICASE|REG_NOSUB);


    for(i=0; i<window; i++)
    {
	GList* events = get_events(date);
	gboolean printed_date = FALSE;

	if(events != NULL)
	{
	    GList* item = g_list_first(events);
	    for(j=0; j<g_list_length(events); j++)
	    {

		if(regexec(&preg, ((PalEvent*) (item->data))->text, 0, NULL, 0)==0 ||
		   regexec(&preg, ((PalEvent*) (item->data))->type, 0, NULL, 0)==0)
		{
		    if(!printed_date && !settings->compact_list)
		    {
			pal_output_date_line(date);
			printed_date = TRUE;
		    }

		    search_hits++;
		    pal_output_event((PalEvent*) (item->data), date);
		}

		item = g_list_next(item);
	    }
	}

	/* put a blank line between dates that we printed */
	if(printed_date)
	    g_print("\n");

	if(settings->reverse_order)
	    g_date_subtract_days(date, 1);
	else
	    g_date_add_days(date, 1);

    }

    /* no extra newlines when using compact list, so add one here */
    if(settings->compact_list)
	g_print("\n");

    pal_output_attr(BRIGHT, _("[ End search results: %s ] [ %d %s found ]\n"),
		    search, search_hits, search_hits == 1 ? "event" : "events");

    regfree(&preg);
}


/* prints the events on the dates from the starting_date to
 * starting_date+window */
void view_range(GDate* starting_date, gint window)
{
    gint i;

    if(settings->reverse_order)
	g_date_add_days(starting_date,window-1);

    for(i=0; i<window; i++)
    {
	pal_output_date(starting_date,FALSE);

	if(settings->reverse_order)
	    g_date_subtract_days(starting_date,1);
	else
	    g_date_add_days(starting_date,1);
    }
}



/* Returns a GDate object for the given key (in_string) */
/* This function only checks the one-time date events (not todo, or
 * recurring events).  It returns NULL on a failure and can optionally
 * print an error message on failure. */
GDate* get_query_date(gchar* in_string, gboolean show_error)
{
    GDate* to_show = NULL;
    gchar* date_string = g_ascii_strdown(in_string, -1);

    if(date_string == NULL)
	return NULL;

    g_strstrip(date_string);

    to_show = g_date_new();
    g_date_set_time(to_show, time(NULL));

    /* these could be better... */
    if(strncmp(date_string, "tomorrow", strlen("tomorrow")) == 0)
    {
	g_date_add_days(to_show, 1);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "yesterday", strlen("yesterday")) == 0)
    {
	g_date_subtract_days(to_show, 1);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "today", strlen("today")) == 0)
    {
	g_date_subtract_days(to_show, 0);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "mo", strlen("mo")) == 0 ||
       strncmp(date_string, "next mo", strlen("next mo")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 1);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "tu", strlen("tu")) == 0 ||
       strncmp(date_string, "next tu", strlen("next tu")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 2);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "we", strlen("we")) == 0 ||
       strncmp(date_string, "next we", strlen("next we")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 3);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "th", strlen("th")) == 0 ||
       strncmp(date_string, "next th", strlen("next th")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 4);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "fr", strlen("fr")) == 0 ||
       strncmp(date_string, "next fr", strlen("next fr")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 5);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "sa", strlen("sa")) == 0 ||
       strncmp(date_string, "next sa", strlen("next sa")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 6);
	return to_show;
    }
    if(strncmp(date_string, "su", strlen("su")) == 0 ||
       strncmp(date_string, "next su", strlen("next su")) == 0)
    {
	do g_date_add_days(to_show, 1); while(g_date_get_weekday(to_show) != 7);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last mo", strlen("last mo")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 1);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last tu", strlen("last tu")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 2);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last we", strlen("last we")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 3);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last th", strlen("last th")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 4);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last fr", strlen("last fr")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 5);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last sa", strlen("last sa")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 6);
	g_free(date_string);
	return to_show;
    }
    if(strncmp(date_string, "last su", strlen("last su")) == 0)
    {
	do g_date_subtract_days(to_show, 1); while(g_date_get_weekday(to_show) != 7);
	g_free(date_string);
	return to_show;
    }


    if(g_ascii_isdigit(*(date_string))) /* if it begins with a digit */
    {
	gchar* ptr = date_string;
	gint date_offset = 0;

	while(g_ascii_isdigit(*ptr))
	    ptr++;

	/* if there were characters before end of the string */
	if(*ptr != '\0')
	{
	    sscanf(date_string, "%d", &date_offset);

	    if(strncmp(ptr, " days away", strlen(" days away")) == 0)
	    {
		g_date_add_days(to_show, date_offset);
		g_free(date_string);
		return to_show;
	    }

	    if(strncmp(ptr, " days ago", strlen(" days ago")) == 0)
	    {
		g_date_subtract_days(to_show, date_offset);
		g_free(date_string);
		return to_show;
	    }
	}

	else /* all digits were used */
	{
	    gint query_date_int = -1;

	    sscanf(date_string, "%d", &query_date_int);


	    /* check for 0, negative number, or excessive preceeding 0's */
	    if(!(query_date_int <= 0 ||
		(*date_string == '0' && *(date_string+1) == '0' &&
		 *(date_string+2) == '0' && *(date_string+3) == '0') ||
		(*(date_string+4) == '0' && *(date_string+5) == '0') ||
		(*(date_string+6) == '0' && *(date_string+7) == '0')))
	    {

		if(query_date_int < 32)
		{
		    if(query_date_int < g_date_get_day(to_show))
			g_date_add_months(to_show,1);

		    if(g_date_valid_dmy(query_date_int,
					(GDateMonth) g_date_get_month(to_show),
					(GDateYear) g_date_get_year(to_show)))
		    {
			g_date_set_day(to_show, query_date_int);
			g_free(date_string);
			return to_show;
		    }
		}
		else if(query_date_int < 1232)
		{
		    gint day, month;

		    day = query_date_int % 100;
		    month = query_date_int / 100;

		    if(day > 0 && day < 32 && month > 0 && month < 13)
		    {
			if(month < g_date_get_month(to_show))
			    g_date_add_years(to_show,1);
			if(day   < g_date_get_day(to_show) && month == g_date_get_month(to_show))
			    g_date_add_years(to_show, 1);

			if(g_date_valid_dmy((GDateDay) day,
					    (GDateMonth) month,
					    (GDateYear) g_date_get_year(to_show)))
			{
			    g_date_set_dmy(to_show, (GDateDay) day,
					   (GDateMonth) month,
					   (GDateYear) g_date_get_year(to_show));
			    g_free(date_string);
			    return to_show;
			}
		    }
		}
		else if(query_date_int > 10000)
		{
		    gint day, month, year;

		    day = query_date_int % 100;
		    month = (query_date_int % 10000 - day) / 100;
		    year = query_date_int / 10000;

		    if(day > 0 && day < 32 && month > 0 && month < 13 && year > 0)
		    {
			if(g_date_valid_dmy((GDateDay) day,
					    (GDateMonth) month,
					    (GDateYear) year))
			{
			    g_date_set_dmy(to_show, (GDateDay) day,
					   (GDateMonth) month,
					   (GDateYear) year);
			    g_free(date_string);
			    return to_show;
			}
		    }
		}
	    }
	}
    }

    /* glib is last resort, but don't let a few things get to it... */
    if(!((*date_string == '0' && *(date_string+1) == '0' &&
	  *(date_string+2) == '0' && *(date_string+3) == '0') ||
	 (*(date_string+4) == '0' && *(date_string+5) == '0') ||
	 (*(date_string+6) == '0' && *(date_string+7) == '0')))
    {
	g_date_set_parse(to_show, date_string);

	if(g_date_valid(to_show))
	{
	    g_free(date_string);
	    return to_show;
	}
    }

    if(show_error)
    {
	/* if we got here, there was an error */
	pal_output_error(_("ERROR: The following date is not valid: %s\n"), date_string);
	pal_output_error(_("       Valid date formats include:\n"));
	pal_output_error(_("       dd, mmdd, yyyymmdd, 'yesterday', 'today', 'tomorrow',\n"));
	pal_output_error(_("       'n days away', 'n days ago',\n"));
	pal_output_error(_("       first two letters of weekday, 'last ' followed by\n"));
	pal_output_error(_("       first two letters of weekday,\n"));
	pal_output_error(_("       '1 Jan 2000', 'Jan 1 2000', etc.\n"));
    }
    g_date_free(to_show);
    g_free(date_string);
    return NULL;
}


/* determines what should be dispalyed if -r, -s, -d are used */
void view_details()
{
    GDate* to_show = settings->query_date;


    if(settings->search_string != NULL &&
       settings->range_days == 0 &&
       settings->range_neg_days == 0)
    {
	pal_output_error(_("ERROR: When using -s, use -r to specify the range to search within.\n"));
    }

    /* if -r and -s isn't used */
    if(settings->range_days == 0 &&
       settings->range_neg_days == 0 &&
       settings->search_string == NULL)
    {
	/* if -d is used, show that day.  Otherwise, show nothing */
	if(to_show != NULL)
	    pal_output_date(to_show, TRUE);
    }

    /* if -r or -s is used, show range of dates relative to -d */
    else if(settings->range_days     != 0 ||
	    settings->range_neg_days != 0)
    {
	GDate* starting_date = NULL;

	if(to_show == NULL) /* if -d isn't used, start from current date */
	{
	    starting_date = g_date_new();
	    g_date_set_time(starting_date, time(NULL));
	}
	else /* otherwise, start from date specified */
	    starting_date = g_memdup(to_show, sizeof(GDate));

	g_date_subtract_days(starting_date, settings->range_neg_days);

	if(settings->search_string == NULL)
	    view_range(starting_date, settings->range_neg_days + settings->range_days);
	else
	    view_search(starting_date, settings->range_neg_days + settings->range_days);

	g_date_free(starting_date);
    }

}



gint parse_arg(gchar** args, gint on_arg, gint total_args)
{
    args = args + on_arg;
    on_arg++;

    if(strcmp(*args,"-h") == 0 ||
       strcmp(*args,"--help") == 0)
    {

	g_print("%s %s - %s\n", _("pal"), VERSION, _("Copyright (C) 2003, Scott Kuhl"));
	g_print(_("  pal is licensed under the GNU General Public License and has NO WARRANTY.\n\n"));
	g_print(_(" -d date      Show events on the given date. (formats: dd, mmdd, yyyymmdd)\n"));
	g_print(_(" -r n         Display events within n days after today or a date used\n"));
	g_print(_("                  with -d. (default: n=0, show nothing)\n"));
	g_print(_(" -r p-n       Display events within p days before and n days after today\n"));
	g_print(_("                  or a date used with -d.\n"));
	g_print(_(" -s regex     Search for events matching the regular expression.\n"));
	g_print(_("                  Use -r to select range of days to search.\n"));
	g_print(_(" -x n         Expunge events that are n or more days old.\n"));
	g_print(_(" -c n         Display calendar with n lines. (default: 5)\n"));
	g_print(_(" -f file      Load 'file' instead of ~/.pal/pal.conf.\n"));
	g_print(_(" -u username  Load /home/username/.pal/pal.conf\n"));
	g_print(_(" -p palfile   Load *.pal file only (overrides files loaded from pal.conf)\n"));
	g_print(_(" -a           Add an event to a file interactively.\n"));
	g_print(_(" --color      Force colors, regardless of terminal type.\n"));
	g_print(_(" --nocolor    Force no colors, regardless of terminal type.\n"));
	g_print(_(" --mail       Generate output readable by sendmail.\n"));
	g_print(_(" --html       Generate html table.  Set size of calendar with -c.\n"));
	g_print(_(" -v           Verbose output.\n"));
	g_print(_(" --version    Display version information.\n"));
	g_print(_(" -h, --help   Display this help message.\n\n"));
	g_print(_("Type \"man pal\" for more information.\n"));
	exit(0);
    }

    if(strcmp(*args,"-r") == 0)
    {
	args++; on_arg++;

	if(on_arg > total_args   || (sscanf(*args, "%d", &(settings->range_days)) != 1 &&
				     sscanf(*args, "%d-%d", &(settings->range_neg_days),
					    &(settings->range_days)) != 2))
	{
	    settings->range_days = 0;
	    settings->range_neg_days = 0;

	    pal_output_error("%s\n", _("ERROR: Number required after -r argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    return on_arg-1;
	}

	settings->range_arg = TRUE;

	if(sscanf(*args, "%d-%d", &(settings->range_neg_days),
		  &(settings->range_days)) == 2)
	    return on_arg;

	if(sscanf(*args, "%d", &(settings->range_days)) == 1)
	{
	    settings->range_neg_days = 0;
	    return on_arg;
	}

	return on_arg;
    }

    if(strcmp(*args,"-c") == 0)
    {
	args++; on_arg++;
	if(on_arg > total_args || sscanf(*args, "%d", &(settings->cal_lines)) != 1)
	{
	    pal_output_error("%s\n", _("ERROR: Number required after -c argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    on_arg--;
	}
	return on_arg;
    }

    if(strcmp(*args,"-d") == 0)
    {
	args++; on_arg++;
	if(on_arg > total_args)
	{
	    pal_output_error("%s\n", _("ERROR: Date required after -d argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    on_arg--;
	}
	else
	{
	    settings->query_date = get_query_date(*args, TRUE);
	    if(settings->query_date == NULL)
		pal_output_error(_("NOTE: Use quotes around the date if it has a spaces.\n"));
	}
	return on_arg;
    }

    if(strcmp(*args,"-v") == 0)
    {
	settings->verbose = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"-a") == 0)
    {
	settings->add_event = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"--color") == 0)
    {
	set_colorize(1);
	return on_arg;
    }

    if(strcmp(*args,"--nocolor") == 0)
    {
	set_colorize(0);
	return on_arg;
    }

    if(strcmp(*args,"--mail") == 0)
    {
	set_colorize(-2); /* overrides a later --color argument */
	settings->mail = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"--html") == 0)
    {
	settings->html_out = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"--version") == 0)
    {
	g_print("pal %s\n", VERSION);
	g_print("Compiled with prefix: %s\n", PREFIX);
	exit(0);
    }

    if(strcmp(*args,"-f") == 0)
    {
	gchar tmp[16384];
	args++; on_arg++;
	if(on_arg > total_args || sscanf(*args, "%s", tmp) != 1)
	{
	    pal_output_error("%s\n", _("ERROR: Pal conf file required after -f argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    return on_arg;
	}

	g_free(settings->conf_file);
	settings->conf_file = g_strdup(tmp);
	settings->specified_conf_file = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"-p") == 0)
    {
	gchar tmp[16384];
	args++; on_arg++;
	if(on_arg > total_args || sscanf(*args, "%s", tmp) != 1)
	{
	    pal_output_error("%s\n", _("ERROR: *.pal file file required after -p argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    return on_arg;
	}

	settings->pal_file = g_strdup(tmp);
	return on_arg;
    }

    if(strcmp(*args,"-u") == 0)
    {
	gchar username[256];
	args++; on_arg++;
	if(on_arg > total_args || sscanf(*args, "%s", username) != 1)
	{
	    pal_output_error("%s\n", _("ERROR: Username required after -u argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    return on_arg;
	}

	g_free(settings->conf_file);
	settings->conf_file = g_strconcat("/home/", username, "/.pal/pal.conf", NULL);
	settings->specified_conf_file = TRUE;
	return on_arg;
    }

    if(strcmp(*args,"-s") == 0)
    {
	args++; on_arg++;

	if(on_arg > total_args)
	{
	    pal_output_error("%s\n", _("ERROR: Regular expression required after -s argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	}
	else
	    settings->search_string = g_strdup(*args);

	return on_arg;
    }

    if(strcmp(*args,"-x") == 0)
    {
	args++; on_arg++;
	if(on_arg > total_args || sscanf(*args, "%d", &(settings->expunge)) != 1)
	{
	    pal_output_error("%s\n", _("ERROR: Number required after -x argument."));
	    pal_output_error("       %s\n", _("Use --help for more information."));
	    on_arg--;
	}

	return on_arg;
    }


    pal_output_error("%s %s\n", _("ERROR: Bad argument:"), *args);
    pal_output_error("       %s\n", _("Use --help for more information."));

    return on_arg+1;
}



/* used when the pal is finished running and the hashtable entries
 * need to be properly freed */
void hash_table_free_item(gpointer key, gpointer value, gpointer user_data)
{
    GList* list = (GList*) value;
    g_free(key);

    /* free the list for this hashtable key */
    while(g_list_length(list) != 0)
    {
	GList* prev = list;
	pal_event_free(list->data);
	list = g_list_next(list);
	g_list_free(prev);
    }
}


int main( gint argc, gchar** argv)
{
    gint on_arg = 1;
    GDate* date  = g_date_new();
    GDate* today = g_date_new();

    g_date_set_time(date,  time(NULL));
    g_date_set_time(today, time(NULL));

    textdomain("pal");

    if(setlocale(LC_MESSAGES, "") == NULL ||
       setlocale(LC_TIME, "") == NULL)
	pal_output_error("WARNING: Localization failed.\n");

    settings = g_malloc(sizeof(Settings));
    settings->cal_lines           = 5;
    settings->range_days          = 0;
    settings->range_neg_days      = 0;
    settings->range_arg           = FALSE;
    settings->search_string       = NULL;
    settings->verbose             = FALSE;
    settings->mail                = FALSE;
    settings->query_date          = NULL;
    settings->expunge             = -1;
    settings->date_fmt            = g_strdup("%a %e %b %Y");
    settings->week_start_monday   = FALSE;
    settings->reverse_order       = FALSE;
    settings->cal_on_bottom       = FALSE;
    settings->specified_conf_file = FALSE;
    settings->no_columns          = FALSE;
    settings->hide_event_type     = FALSE;
    settings->add_event           = FALSE;
    settings->event_color         = BLUE;
    settings->pal_file            = NULL;
    settings->html_out            = FALSE;
    settings->compact_list        = FALSE;
    settings->compact_date_fmt    = g_strdup("%m/%d/%Y");
    settings->conf_file = g_strconcat(g_get_home_dir(), "/.pal/pal.conf", NULL);

    /* figure out the terminal width */
    {
	struct winsize wsz;
	if(ioctl(0, TIOCGWINSZ, &wsz) != -1)
	    settings->term_cols = wsz.ws_col;
	else
	    /* assume 80 columns if unsure...I don't think this will
	     * cause any problems */
	    settings->term_cols = 80;
    }

    /* parse all the arguments */
    while(on_arg < argc)
	on_arg = parse_arg(argv,on_arg,argc);

    ht = load_files();

    /* adjust settings if --mail is used */
    if(settings->mail)
    {
	gchar pretty_date[128];
	g_date_strftime(pretty_date, 128, settings->date_fmt, today);

	g_print("From: \"pal\" <pal>\n");
	g_print("Subject: [pal] %s\n\n", pretty_date);

	/* let the mail reader handle the line wrapping */
	settings->term_cols = -1;
    }
    else if(settings->term_cols < 77)
	/* force a one column calendar if terminal is too small for
	 * two columns to display properly and --mail isn't used */
	    settings->no_columns = TRUE;

    if(!settings->html_out)
    {

	if(!settings->cal_on_bottom)
	{
	    pal_output_cal(date,settings->cal_lines,today);
	    /* print a newline under calendar if we're printing other stuff */
	    if(settings->cal_lines > 0 && (settings->range_days>0 ||
					   settings->range_neg_days>0 ||
					   settings->query_date != NULL))

		g_print("\n");

	}

	view_details(); /* prints results of -d,-r,-s */

	if(settings->cal_on_bottom)
	{
	/* print a new line over calendar if we've printed other stuff */
	    if(settings->cal_lines > 0 && (settings->range_days>0 ||
					   settings->range_neg_days>0 ||
					   settings->query_date != NULL))
		g_print("\n");

	pal_output_cal(date,settings->cal_lines,today);
	}


	if(settings->add_event)
	{
	    /* this is always last so we always need a newline here */
	    g_print("\n");
	    pal_add_event();
	}
    } /* end if html_out */
    else
	pal_html_out();

    g_date_free(today);
    g_date_free(date);

    /* free hash table */
    g_hash_table_foreach(ht, (GHFunc) hash_table_free_item, NULL);
    g_hash_table_destroy(ht);

    /* free settings */
    g_free(settings->date_fmt);
    g_free(settings->conf_file);
    g_free(settings->search_string);
    if(settings->query_date != NULL)
	g_date_free(settings->query_date);
    g_free(settings->compact_date_fmt);
    g_free(settings->pal_file);

    g_free(settings);

    return 0;
}
